local entity = require("entity")
local hash = require("hash")
local weapon = require("weapon")

local abilities = {}

local data = nil

local function randomDirection(ability)
    local dir = math.random() * 2 * math.pi
    return math.cos(dir), math.sin(dir), dir
end

local function playerDir(ability, gameData)
    return gameData.player.dir, 0
end

local function enemiesWithinRadiusSorted(gameData, radius)
    local player = gameData.player
    local enemies = gameData.enemies
    local enemiesHash = gameData.enemiesHash

    local results, count = hash.search(enemiesHash, player.x, player.y, radius)
    local targets = {}

    for i = 1, count do
        local e = enemies[results[i]]
        local dist  = math.dist(player.x, player.y, e.x, e.y)
        if dist < radius + e.cr then
            table.insert(targets, {e, dist})
        end
    end

    table.sort(targets, function(a, b)
        return a[2] < b[2]
    end)

    return targets
end

function makeProjectile(def, dx, dy, rotation, gameData)
    local player = gameData.player
    local stats = gameData.playerStats
    local proj = entity.make(def.prefab)

    proj.x = player.x
    proj.y = player.y
    proj.rotation = rotation + def.startAngle
    proj.vx = dx * def.speed * stats.attackSpeedModifier
    proj.vy = dy * def.speed * stats.attackSpeedModifier
    proj.power = def.power * stats.attackPowerModifier
    proj.liveTime = def.liveTime * stats.attackLifetimeModifier

    return proj
end

function nearestTarget(def, gameData, addProjectile)
    local player = gameData.player
    local stats = gameData.playerStats
    local playerbuls = gameData.playerbuls

    
    local targets = enemiesWithinRadiusSorted(gameData, def.radius * stats.targetRadiusModifier)

    for i = 1, math.min(#targets, def.count) do
        local e = unpack(targets[i])
        local dx, dy = math.norm(e.x - player.x, e.y - player.y)
        local angle = math.angle(dx, dy) 
        local proj = makeProjectile(def, dx, dy, angle, gameData)
       
        addProjectile(proj)
    end
end

function abilities.init()
    data = {}

    data.arrow = abilities.make({
        icon = "powerup_arrow",
        cooldownTime = 0.8,
        power = 1,
        count = 1,
        liveTime = 0.5,
        speed = 180,
        startAngle = 0.5 * math.pi,
        radius = 64,
        prefab = {
            cx = 0, cy = 0, cr = 4,
            sprite = "weapon_arrow",
            update = weapon.updateProjectile
        },
        fire = nearestTarget

    })
end


function abilities.update(dt, abilityName, gameData)
    local stats = gameData.playerStats
    local obj = data[abilityName]

    function addProjectile(proj)
        table.insert(gameData.playerbuls, proj)
    end
    if obj then
        obj.timer = obj.timer - dt * stats.attackCooldownModifier
        if obj.timer <= 0 then
            
            obj.timer = obj.timer + obj.cooldownTime
            if obj.fire then
                obj:fire(gameData, addProjectile)
            end
        end
    end
end

function abilities.make(prot)
    local ability = entity.make(prot)
    ability.level = 1
    ability.timer = ability.cooldownTime

    return ability
end

function abilities.getAbility(abilityName)
    return data[abilityName]
end

function abilities.applyUpgrade(abilityName, upgradeData)
    local obj = data[abilityName]
    local field = upgradeData.field
    if obj and obj[field] then
        obj[field] = upgradeData.apply(obj[field])
    end
end



return abilities