local Signal = require("lib.hump.signal")
local input = require("input")
local drawing = require("drawing")
local music = require("music")
local assets = require("assets")
local collision = require("collision")
local player = require("player")
local entity = require("entity")
local enemy = require("enemy")
local floor = require("floor")
local hud = require("hud")
local powerups = require("powerups")
local powerupmenu = require("ui.powerupmenu")
local weapon = require("weapon")
local state = require("state")

local transitionTime = 40
uiFrame = false

local slimeFragment = {
    lifetime = 10,
    layer = "collectable",
    sprite = "slimeFragment",
    cx = 0, cy = 0,
    w = 12, h  = 9
}

local gameData

local game = {}

local frame = 0

local waves = {
    {
        totalEnemies = 20,
        concurrentEnemies = 6,
        minSpawnInterval = 1,
        maxSpawnInterval = 2 
    },
    {
        totalEnemies = 25,
        concurrentEnemies = 8,
        minSpawnInterval = 1,
        maxSpawnInterval = 2 
    },
    {
        totalEnemies = 30,
        concurrentEnemies = 8,
        minSpawnInterval = 1,
        maxSpawnInterval = 2 
    },
    {
        totalEnemies = 35,
        concurrentEnemies = 8,
        minSpawnInterval = 0.8,
        maxSpawnInterval = 1.5 
    },
    {
        totalEnemies = 40,
        concurrentEnemies = 12,
        minSpawnInterval = 0.8,
        maxSpawnInterval = 1.5
    },
    {
        totalEnemies = 50,
        concurrentEnemies = 12,
        minSpawnInterval = 0.8,
        maxSpawnInterval = 1.5
    },
    {
        totalEnemies = 50,
        concurrentEnemies = 12,
        minSpawnInterval = 0.5,
        maxSpawnInterval = 1.2
    },
    {
        totalEnemies = 60,
        concurrentEnemies = 15,
        minSpawnInterval = 0.5,
        maxSpawnInterval = 1.2
    },
    {
        totalEnemies = 60,
        concurrentEnemies = 15,
        minSpawnInterval = 0.5,
        maxSpawnInterval = 1.2
    },
    {
        totalEnemies = 80,
        concurrentEnemies = 17,
        minSpawnInterval = 0.25,
        maxSpawnInterval = 1 
    },
} 

local function nearComputer()
    local x = gameData.reactor.x
    local y = gameData.reactor.y + 32
    local w = 32
    local h = 8

    return collision.bodyOverlaps(gameData.player, x, y, w, h)
end


function game.init()
    love.mouse.setVisible(false)
    frame = 0
    
    collision.init()

    gameData = {
        currency = 500,
        state = "playing",
        player = player.make(WIDTH / 2, HEIGHT / 2 + 32),
        playerHelpers = {},
        gunX = math.round(WIDTH / 2) , 
        gunY = math.round(HEIGHT / 2 + 32),
        gunFlip = false,
        reactor = entity.make({
            sprite = "shrine", 
            name = "reactor",
            x = WIDTH / 2 - 16,
            y = 8 * 16,
            w = 32, h = 32,
            solid = true, 
            hp = 15,
            maxHp = 15,
            efficiency = 1
        }),
        level = 1,
        experience = 0,
        experienceToNextLevel = 2,
        playerbuls = {},
        slimeFragments = {},
        playerStats = {
            movementSpeedModifier = 1,
            targetRadiusModifier = 1,
            attackLifetimeModifier = 1,
            attackPowerModifier = 1,
            attackCooldownModifier = 1, 
            attackSpeedModifier = 1,
        },
        abilities = {"arrow"},
        enemies = {},
        pickups = {},     
        cx = 0,
        cy = 0,

        wave = 0,
        waveTimer = 0,
        waveCounter = 0,
        waveStarted = false,
        wavesSurvived = 0,
        drawList = {},
        drawCount = 0
    }  

    hud.init(gameData)
end

function game.getReactor()
    return gameData.reactor
end

function game.getPlayer()
    return gameData.player
end

function game.subtractCurrency(amount)
    gameData.currency = math.max(0, gameData.currency - amount)
end


function game.hasAMouse()
    local helpers = gameData.playerHelpers
    for i = 1, #helpers do
        if helpers[i].type == "mouse" then
            return true
        end
    end
    return false
end

function game.startNextWave() 
    gameData.waveStarted = true
    gameData.waveComplete = false
    gameData.wavesSurvived = gameData.wavesSurvived + 1
    gameData.wave = math.min(gameData.wave + 1, #waves)
    gameData.waveCounter = 0
    gameData.waveTimer = 0
    music.fadeIn("waveMusic")
end

function game.waveInProgress()
    return #gameData.enemies > 0 and gameData.waveCounter <= waves[gameData.wave].totalEnemies
end

function game.getCurrency()
    return gameData.currency
end

function game.pause()
    love.mouse.setVisible(true)
end

function game.resume()
    love.mouse.setVisible(false)
end

function game.addPlayerBullet(prototype, x, y)
    local e = entity.make(prototype, x, y)
    table.insert(gameData.playerbuls, e)
end

function game.giveAbility(abilityName)
    table.insert(gameData.abilities, abilityName)
end

function game.getPlayerStats()
    return gameData.playerStats
end

local function addToDrawList(obj)
    local list = gameData.drawList
    local count = gameData.drawCount + 1

    list[count] = obj
    gameData.drawCount = count
    
end



local function updatePlayerBullets(dt)
    for i = #gameData.playerbuls, 1, -1 do
        local bul = gameData.playerbuls[i]
        entity.update(dt, bul, gameData)
        bul.rotation = math.angle(bul.vx, bul.vy) + 0.5 * math.pi

        if bul.destroy or collision.bodyOverlaps(bul, 0, 0, WIDTH, HEIGHT) == false then
            table.remove(gameData.playerbuls, i)
            collision.removeBody(bul)
        end  
    end
end

local function updatePlayerHelpers(dt)
    for i = #gameData.playerHelpers, 1, -1 do
        local helper = gameData.playerHelpers[i]
        entity.update(dt, helper, gameData) 

        if helper.destroy or helper.hp <= 0 then
            table.remove(gameData.playerHelpers, i)
            collision.removeBody(helper)
        end
    end
end



local function updateEnemies(dt)
    local player = gameData.player
    local abilities = gameData.abilities
    local playerbuls = gameData.playerbuls
    local enemies = gameData.enemies
    local slimeFragments = gameData.slimeFragments

 
    for i = #enemies, 1, -1 do
        local e = enemies[i]
        if e.hp <= 0 then
            table.remove(enemies, i)
            collision.removeBody(e)
            
            table.insert(gameData.slimeFragments, entity.make(slimeFragment, e.x, e.y))
        else
            entity.update(dt, e, gameData)
        end
    end

   
    -- TODO this using collision.query 
    -- for i = 1, #enemies do
    --     if i % 3 == frame % 3 then
    --         local e1 = enemies[i]
            
    --         local res, c = hash.search(gameData.enemiesHash, e1.x, e1.y, 4)
            
    --         for j = 1, c do     
    --             local idx = res[j]
    --             if idx ~= i then
                    
    --                 local e2 = enemies[idx]
    --                 local d = entity.distance(e1, e2)
    --                 local r = e1.cr + e2.cr
    --                 if d < r then
    --                     local push = r - d
    --                     local dx, dy = math.norm(e2.x - e1.x, e2.y - e1.y)
    --                     e1.x = e1.x - (dx * push * 0.5)
    --                     e1.y = e1.y - (dy * push * 0.5)
    --                     e2.x = e2.x + (dx * push * 0.5)
    --                     e2.y = e2.y + (dy * push * 0.5)
    --                 end
    --             end
    --         end
    --     end
    -- end

end

local function updateWave(dt)
     if gameData.waveStarted and not gameData.waveComplete then
        local wave = waves[gameData.wave]

        gameData.waveTimer = gameData.waveTimer - dt
        if gameData.waveCounter < wave.totalEnemies then
            if gameData.waveTimer <= 0 and #gameData.enemies < wave.concurrentEnemies then
                local angle = love.math.random() * 2 * math.pi
                local dx, dy = math.cos(angle), math.sin(angle)
                
                local ox = math.min(math.abs(dx * 280), 16 + WIDTH /2 ) * math.sign(dx)
                local oy = math.min(math.abs(dy * 280), 16 + HEIGHT /2) * math.sign(dy)

                local x = (WIDTH / 2) + ox
                local y = (HEIGHT / 2) + oy
                local en = enemy.make("thing",  x, y)
                table.insert(gameData.enemies, en)
                gameData.waveCounter = gameData.waveCounter + 1
                gameData.waveTimer = love.math.random(wave.minSpawnInterval, wave.maxSpawnInterval)
            end   
        else
            if #gameData.enemies == 0 then
                gameData.waveComplete = true 
                music.fadeOut()
            end
        end
    end
end

local function resetDrawList() 
    for i = 1, #gameData.drawList do
        gameData.drawList[i] = nil
    end
    gameData.drawCount = 0
end

local function rebuildDrawList()
    -- add everything within camera bounds and sort along the y-axis
    resetDrawList()

    local tlx, tly = gameData.cx - WIDTH / 2, gameData.cy - HEIGHT / 2 
    local results, count = collision.query(0, 0, WIDTH, HEIGHT)
    for i = 1, count do
        local obj = results[i]
        addToDrawList(obj)
    end

    table.sort(gameData.drawList, function(a,b) 
        return a and b and a.y < b.y
    end)
end

function game.addPlayerHelper(prototype, x, y)
    local e = entity.make(prototype, x, y)
    table.insert(gameData.playerHelpers, e)
    rebuildDrawList()
end

local function simulateStep(dt)
    frame = (frame + 1) % 1000
    if frame % 26 == 0 then
        uiFrame = not uiFrame
    end
    local st = love.timer.getTime()

    updateWave(dt)
    
    local player = gameData.player
   
    local playerbuls = gameData.playerbuls
    local enemies = gameData.enemies
    local fragments = gameData.slimeFragments

    entity.update(dt, player, gameData)

    local obj = player
    local items, count = collision.query(obj.x + obj.cx - obj.cr, obj.y + obj.cy - obj.cr, obj.cr * 2, obj.cr * 2, function(item)
        return item.layer == "collectable"
    end)

    for i = 1, count do
        local item = items[i]
        item.destroy = true
        gameData.currency = gameData.currency + (100 * gameData.reactor.efficiency)
    end

    for i = #fragments,1, -1 do
        local frag = fragments[i]
        frag.lifetime = frag.lifetime - dt
        if frag.lifetime < 2 then
            local ltms = math.floor(frag.lifetime * 1000)
            if ltms % 2 == 0 then
                frag.visible = not frag.visible
            end
        end
        if frag.destroy or frag.lifetime <= 0 then
            table.remove(fragments, i)
            collision.removeBody(frag)
        end
    end


    local mx, my = input.pointerPosition()
    local dx, dy = math.norm(mx - player.x, my - player.y)
    local gunX, gunY = math.round(player.x)  + (dx * 6), math.round(player.y) + dy * 6
    local gunAngle = math.angle(dx, dy) + 0.5 * math.pi
    gameData.gunX = gunX
    gameData.gunY = gunY
    gameData.gunAngle = gunAngle
    gameData.gunFlip = dx < 0
    if input.justPressed("pointer") then
        game.addPlayerBullet({
            liveTime = 1,
            target = "enemies",
            
            
            vx = dx * 256, vy = dy * 256,
            sprite = "demo_bullet",
            cx = 0, cy = 0, cr = 2,
            update = weapon.updateProjectile,
            power = 1,
            projectile = true
        }, gunX + dx * (21 - 8),  gunY + dy * (21 - 8))

        -- table.insert(playerbuls, b)
    end

    updatePlayerBullets(dt)
    updatePlayerHelpers(dt)
    updateEnemies(dt)
 
    rebuildDrawList()
end


function game.update(dt)  
    if input.justPressed("start") then
        state.push("paused")
    end

    if not game.waveInProgress() and  nearComputer() and input.justPressed("interact") then
       state.push("compmenu", true)
    end


    simulateStep(dt)
    hud.update(dt, gameData)

    if gameData.reactor.hp <= 0 then
        state.push("gameover", "reactor")
        music.fadeOut()
    elseif gameData.player.hp <= 0 then
        state.push("gameover", "player")
        music.fadeOut()
    end
    
end

function game.draw()
    local player = gameData.player
    local mx, my = input.pointerPosition()

    drawing.cls(1,0,0,1)

    drawing.camera(gameData.cx, gameData.cy)

    floor.draw(gameData.cx, gameData.cy)

    for i = 1, gameData.drawCount do
        local obj = gameData.drawList[i]
        entity.draw(obj)
    end

    drawing.sprite("weapon_arrowgun", gameData.gunX, gameData.gunY , gameData.gunAngle, gameData.gunFlip)

    drawing.camera()

    hud.draw(gameData)

    if not game.waveInProgress() and nearComputer() then
        local spr = "uiKeyEReleased"
        if uiFrame then
            spr = "uiKeyEPressed"
        end
        drawing.sprite(spr, player.x - 8, player.y - 32)
        -- drawing.printf("Press E", 0, HEIGHT / 2 - 16, WIDTH, "center")
    end
    drawing.col(1, 1, 1, 1)

    if state.current() == "game" then
        drawing.draw(assets.images.cursor, math.round(mx) -7 , math.round(my) -7 )
    end
    
    -- collision.drawDebug()
    if gameData.wave > 0 then
        drawing.print(tostring(gameData.waveCounter) .. "/" .. tostring(waves[gameData.wave].totalEnemies), 2, 80)
    end
end

return game