--[[
More or less a straight copy of https://github.com/matthias-research/pages/blob/master/tenMinutePhysics/11-hashing.html
Hashing function is a bit rubbish because bitwise operators are not supported in Love.js, and the compatiblilty library is 
too slow for my needs, but with a big enough table the collisions are kept reasonably low at the expense of higher memory use
]]
local hash = {}


local function toInt(sh, x, y)
    return math.floor(x / sh.cellSize), math.floor(y / sh.cellSize)
end


local function getHash(sh, x, y)
    local h = x * 8677 - y * 8081  
    return (math.abs(h) % sh.tableSize) + 1
end

local function getHashi(sh, x, y)
    x, y = toInt(sh, x, y)
    return getHash(sh, x, y)
end


function hash.init(_cellSize, _maxElements)
    local cellSize = _cellSize
    local tableSize = _maxElements * 4
    local cellStart = {}
    local cellEntries = {}
    local queryResults = {}
    local seen = {}
    for i = 1, tableSize do
        queryResults[i] = -1
    end

    local res = {
        maxElements = _maxElements, 
        ["cellSize"] = cellSize,
        ["tableSize"] = tableSize,
        ["cellStart"] = cellStart,
        ["cellEntries"] = cellEntries,
        ["queryResults"] = queryResults,
        ["seen"] = seen
    }
    hash.clear(res)
    return res
end

function hash.clear(h)
    for i = 1, h.tableSize do
        h.cellStart[i] =  0
        h.cellEntries[i] =  0
    end
    h.cellStart[h.tableSize + 1] = 0 
end

function hash.update(sh, objs)
    hash.clear(sh)

    local cellStart = sh.cellStart
    local cellEntries = sh.cellEntries

    for i = 1, #objs do
        local o = objs[i]
        local h = getHashi(sh, o.x + (o.cx or 0), o.y + (o.cy or 0))
        cellStart[h] = cellStart[h] + 1
    end

    local start = 0
    for i = 1, sh.tableSize do
        
        start = start + cellStart[i] 
        cellStart[i] = start 
    end

    cellStart[sh.tableSize + 1] = start

    for i = 1, #objs do
        local o = objs[i]
        local h = getHashi(sh, o.x + (o.cx or 0), o.y + (o.cy or 0))
        cellStart[h] = cellStart[h] - 1
        cellEntries[cellStart[h] + 1] = i
    end


end

function hash.search(sh, _x, _y, d)  
    local sx, sy = _x - d, _y - d
    local ex, ey = _x + d, _y + d
    return hash.searchRect(sh, sx, sy, ex, ey)
end
    
function hash.searchRect(sh, sx, sy, ex, ey)
    sx, sy = toInt(sh, sx, sy)
    ex, ey = toInt(sh, ex, ey)

    local cellStart = sh.cellStart
    local cellEntries = sh.cellEntries
    local queryResults = sh.queryResults
    local seen = sh.seen
    local resultsSize = 0

    -- for k, v in pairs(seen) do
    for i = 1, sh.maxElements do
        seen[i] = false
    end

    for y = sy, ey do
        for x = sx, ex do
            local h = getHash(sh, x, y)
            
            local s = cellStart[h]
            local e = cellStart[h + 1]
            
            for i = s, e - 1  do
                local id = cellEntries[i + 1]
                if not seen[id] then
                    queryResults[resultsSize + 1] = cellEntries[i + 1]
                    resultsSize = resultsSize + 1
                    seen[id] = true
                end
            end
        end
    end

    return queryResults, resultsSize
end

return hash